(ns one.todomvc.definitions
  (:use [domina :only [value]])
  (:require [enfocus.core :as ef]
            [clojure.browser.event :as event]
            [goog.events.KeyCodes :as key-codes]
            [goog.events.KeyHandler :as key-handler]
            [one.dispatch :as dispatch]
            [goog.dom.forms :as forms])
  (:require-macros [enfocus.macros :as em]))

;; UTILS
(defn by-id [id] (.getElementById js/document id))

(defn ev-obj [event]
  (.-currentTarget event))

(defn add-input-event-listeners
  [field-id]
  (let [field (by-id field-id)
        keyboard (goog.events.KeyHandler. (by-id field-id))]
    (event/listen field
                  "focus"
                  #(dispatch/fire [:editing-field field-id]))
    (event/listen field
                  "keyup"
                  #(dispatch/fire [:field-changed field-id] (value field)))
    (event/listen keyboard
                  "key"
                  (fn [e] (when (= (.-keyCode e) key-codes/ENTER)
                           (dispatch/fire :change-todo {:title field-id}))))))

(defn add-toggle-listener
  [id]
  (let [field (by-id id)]
    (event/listen field "click" (fn [e] (mark-done e)))))

(defn get-grandfather [node]
  (.-parentNode (.-parentNode node)))

(defn render-link [node]
  (em/at node (em/set-attr :class "selected")))

(defn clear-completed [event]
  (let [node (.-currentTarget event)]
    (em/at (get-grandfather node) ["#main > ul > li.completed"] (em/remove-node)))
  (dispatch/fire :clear-completed))

(defn del-todo [event]
  (let [node (.-currentTarget event)
        label (em/from (em/select (.-parentNode node) ["label"]) (em/get-text))]
    (dispatch/fire :deltodo label)))

(defn do-edit [event]
  (let [node (.-currentTarget event)
        title (em/from (em/select node ["label"]) (em/get-text))
        sel [(str "input" "#" title ".edit")]
        input (em/at node sel)
        todo (em/from js/document :todo sel (em/get-prop :value))]
    (ef/log-debug (str "title=" title " todo=" todo))
    (em/at node (em/set-attr :class "editing"))
    (.-focus input)
))

(defn edit-todo [event]
  (let [new-title (em/from (em/select node ["label"]) (em/get-text))]
    (dispatch/fire :editing-field new-title)))

;; HACK waiting for Creighton macro
(defn set-checked [checked?]
  (ef/chainable-standard
   (fn [node](set! (.-checked node) checked?))))

(defn set-value [v]
  (ef/chainable-standard
   (fn [node] (set! (.-value node) v))))

(defn toggle-all [event]
  (let [node (.-currentTarget event)]
    (dispatch/fire :toggle-all (.-checked node))))

(defn mark-done [event]
  (let [node (.-currentTarget event)
        completed? (.-checked node)
        label (em/from (em/select (.-parentNode node) ["label"]) (em/get-text))]
    (ef/log-debug (str "completed?=" completed?))
    (dispatch/fire :completed {:completed? (not completed?) :title label})))

;; TEMPLATES
(em/deftemplate todoapp "templates/todoapp.html" [])

(em/deftemplate todoitem "templates/todo-li.html" [label completed]
  ["input[name='done']"] (set-checked completed)
  ["#label"] (em/content label)
  ["li > input.edit"] (em/set-attr :value label :id label)
  ["li"](em/set-attr :class (if completed "completed" " "))
  )

;; ACTIONS
(em/defaction start []
  ["#container"](em/content (todoapp))
  ["button#clear-completed"] (em/listen :click clear-completed)
  ["input#toggle-all"] (em/listen :click toggle-all)
  ["#filters > li > a"] (em/listen :click #(em/do->
                                            (reset-class)
                                            (render-link (.-currentTarget %)))))

(em/defaction render [todos count]
  ["#main"] (em/set-attr :style (if (< 0 (:count count)) " " "display: none;"))
  ["input#new-todo"] (set-value nil)
  ["ul#todo-list > li"] (em/remove-node)
  
  ["#todo-list"] (em/append (map #(todoitem (:title %) (:completed? %)) todos))
  
  ["#todo-list button"](em/listen :click del-todo)
  ["#todo-list input[type='checkbox']"] (em/listen :click mark-done)

  ["ul#todo-list"] (ef/chainable-standard
                    (fn [node]
                      (doseq [td todos]
                        (let [title (:title td)
                              completed? (em/at node ["input[type='checkbox']"] (em/get-attr :value))]
                          (ef/log-debug (str "completed?=" completed?))
                          (em/at node ["li"] (em/listen :dblclick do-edit))
                          (add-input-event-listeners title)))))
  
  ["section#todoapp > footer"] (if (= 0 (:count count))
                                 (em/set-attr :style "display: none;")
                                 (em/set-attr :style " "))
  
  ["span strong"] (em/content (str (:left count)))
  ["button#clear-completed"](em/content (str "Clear completed (" (:completed count) ")"))
  )


(em/defaction reset-class []
  ["#filters > li > a"] (em/set-attr :class " "))
